
//-------------------------------------------------------------------------------
// 	2up.jit.centroids-- generate centroids of clumps of intensity from a 1D float matrix.
//  suitable for use with 2up.jit.tactex.
// 
// 	author: randy jones rej@2uptech.com
// 	created 23 Mar 2004
// 	UB version Feb 2007

#include "jit.common.h"
#include <math.h>

// --------------------------------------------------------------------------------
// 

#define	MAX_CENTROIDS	128					// max number of centroids reported
#define	MAX_POSSIBLE_CENTROIDS	1024		// number of potential centroids gathered before sorting.

// e_pixdata-- stores which adjacent pixels are less in intensity. 
// 
typedef unsigned char e_pixdata;
#define	PIX_RIGHT 	0x01
#define	PIX_UP		0x02
#define	PIX_LEFT	0x04
#define	PIX_DOWN	0x08
#define	PIX_ALL		0x0F
#define PIX_DONE	0x80

typedef struct _centroid_info
{
	float					fp_sum;
	float					fx;
	float					fy;
	long					x_sum;
	long					y_sum;
}	t_centroid_info;

t_centroid_info g_zero_centroid = {0., 0., 0., 0, 0};

typedef struct _jit_centroids 
{
	t_object				ob;
	void					*obex; 
	void 					*matrixout;
	void 					*listout;
	float					threshold;				// only include pixels above this value
	long					match;					// match old centroids with new at each frame
	int						width;
	int						height;
	e_pixdata *				mpc_map;
	t_centroid_info	**		p_c_info;
	t_centroid_info **		p_new_centroids;
	short * 				matchfwd;
	short * 				matchback;
	short * 				was_centroid;
	long					max_centroids;
	long					curr_centroids;
	char *					p_in_data;				// ptr to input matrix base addr
	long					in_rowbytes;
} 	t_jit_centroids;


void *jit_centroids_new(t_symbol *s, long argc, t_atom *argv);
void jit_centroids_free(t_jit_centroids *x);
void jit_centroids_assist(t_jit_centroids *x, void *b, long msg, long arg, char *dst);
void free_map(t_jit_centroids *x);

void jit_centroids_jit_matrix(t_jit_centroids *x, t_symbol *s, int argc, t_atom *argv);
void gather_and_report_centroids(t_jit_centroids *x);
void gather_centroid(t_jit_centroids *x, int i, int j, t_centroid_info * c);
#pragma mark -


t_class *jit_centroids_class;

void main(void)
{	
	long attrflags;
	void *classex, *attr;
	
	setup((t_messlist **)&jit_centroids_class, (method)jit_centroids_new, (method)jit_centroids_free, (short)sizeof(t_jit_centroids), 
		0L, A_GIMME, 0);

	classex = max_jit_classex_setup(calcoffset(t_jit_centroids, obex));
	
 	attrflags = JIT_ATTR_GET_DEFER_LOW | JIT_ATTR_SET_USURP_LOW ;

	// attributes	
	attr = jit_object_new(_jit_sym_jit_attr_offset, "threshold", _jit_sym_float32, attrflags,
		(method)0L, (method)0L, calcoffset(t_jit_centroids, threshold));	
	max_jit_classex_addattr(classex, attr);	
	
	attr = jit_object_new(_jit_sym_jit_attr_offset, "max", _jit_sym_long, attrflags,
		(method)0L, (method)0L, calcoffset(t_jit_centroids, max_centroids));	
	max_jit_classex_addattr(classex, attr);	
	
	attr = jit_object_new(_jit_sym_jit_attr_offset, "match", _jit_sym_long, attrflags,
		(method)0L, (method)0L, calcoffset(t_jit_centroids, match));	
	max_jit_classex_addattr(classex, attr);	
	
	max_addmethod_usurp_low((method)jit_centroids_jit_matrix, "jit_matrix");
	
	addmess((method)jit_centroids_assist, "assist", A_GIMME, 0L);
	
	max_jit_classex_standard_wrap(classex,NULL,0);	
    max_jit_class_addmethods(jit_class_findbyname(gensym("2up_jit_centroids")));
}


void *jit_centroids_new(t_symbol *s, long argc, t_atom *argv)
{
	t_jit_centroids *x;
	long i;
	
	x = (t_jit_centroids *)max_jit_obex_new(jit_centroids_class,gensym("2up_jit_centroids"));
	max_jit_obex_dumpout_set(x, outlet_new(x,0L)); 			// general purpose outlet(rightmost)
	x->listout = outlet_new(x,0L); 							// list output

	// init vars
	x->threshold = 1.;
	x->match = 1;
	x->mpc_map = 0;
	x->p_in_data = 0;
	x->in_rowbytes = 0;
	x->width = 0;
	x->height = 0;
	x->curr_centroids = 0;
	x->max_centroids = 8;	// default

	// allocate data for gather_and_report_centroids
	x->p_c_info = jit_getbytes(sizeof(void *) * MAX_CENTROIDS);
	x->p_new_centroids = jit_getbytes(sizeof(void *) * MAX_POSSIBLE_CENTROIDS);
	x->matchfwd = jit_getbytes(sizeof(short) * MAX_CENTROIDS);
	x->matchback = jit_getbytes(sizeof(short) * MAX_CENTROIDS);
	x->was_centroid = jit_getbytes(sizeof(short) * MAX_CENTROIDS);
	
	if ((!x->p_c_info) || (!x->matchfwd) || (!x->matchback) || (!x->was_centroid) || (!x->p_new_centroids)) 
	{
		error("2up.jit.centroids: out of memory!");
		x = 0;
		goto out;
	}
	
	for (i=0;i<MAX_CENTROIDS;i++)
	{
		x->p_c_info[i] = 0;
		x->matchfwd[i] = 0;
		x->matchback[i] = 0;
		x->was_centroid[i] = 0;
	}
	
	for (i=0;i<MAX_POSSIBLE_CENTROIDS;i++)
	{
		x->p_new_centroids[i] = 0;
	}
	
	max_jit_attr_args(x,argc,argv); 						// handle attribute args

out:	
	return (x);
}

void jit_centroids_free(t_jit_centroids *x)
{
	int i;
	// free centroids
	for (i=0; i<MAX_POSSIBLE_CENTROIDS; i++)
	{
		if (x->p_new_centroids[i])
		{
			jit_freebytes(x->p_new_centroids[i], sizeof(t_centroid_info));
		}
	}
	for (i=0; i<MAX_CENTROIDS; i++)
	{
		if (x->p_c_info[i])
		{
			jit_freebytes(x->p_c_info[i], sizeof(t_centroid_info));
		}
	}
	
	jit_freebytes(x->matchfwd, (sizeof(short) * MAX_CENTROIDS));
	jit_freebytes(x->matchback, (sizeof(short) * MAX_CENTROIDS));
	jit_freebytes(x->was_centroid, (sizeof(short) * MAX_CENTROIDS));

	free_map(x);
	max_jit_obex_free(x);
}

void jit_centroids_assist(t_jit_centroids *x, void *b, long msg, long arg, char *dst)
{
	if (msg==1) 
	{
		if (arg == 0)
		{
			sprintf(dst, "matrix in (float32, 1 plane)");
		}
	} 
	else if (msg==2) 
	{
		if (arg == 0)
		{
			sprintf(dst, "centroids: list (index, x, y, intensity)");
		}
		else if (arg == 1)
		{
			sprintf(dst, "dumpout");
		}
	}
}


void free_map(t_jit_centroids *x)
{
	if (x->mpc_map)
	{	
		jit_freebytes(x->mpc_map, x->width * x->height);
		x->mpc_map = 0;
	}
}


void jit_centroids_jit_matrix(t_jit_centroids *x, t_symbol *s, int argc, t_atom *argv)
{

	t_jit_matrix_info 		info;
	t_symbol * 				matrix_name = 0;
	void *					matrix = 0;
	int						i, j;
	float *					pf_data = 0;
	float *					pf_data_prev = 0;
	float *					pf_data_next = 0;
	int						width, height;
	e_pixdata *				p_map, * p_map_prev = 0, * p_map_next = 0;
	int						firstrow, lastrow, firstcol, lastcol;
	float					pixel;
	

	// get matrix
	matrix_name = jit_atom_getsym(&argv[0]);
	if (matrix_name != _jit_sym_nothing) 
	{
		matrix = jit_object_findregistered(matrix_name);
	}
	if (!matrix) 
	{
		error ("2up.jit.centroids: couldn't get matrix object %s!", matrix_name->s_name);
		return;
	}
	
	jit_object_method(matrix,_jit_sym_getinfo,&info);
	jit_object_method(matrix,_jit_sym_getdata,&(x->p_in_data));

	if (x->p_in_data == 0)
	{
		error("2up.jit.centroids: null data ptr for matrix!");
		return;
	}	
	if (info.dimcount != 2)
	{
		error("2up.jit.centroids: input matrix must be 2D.");
		return;
	}	
	if (info.planecount != 1)
	{
		error("2up.jit.centroids: input matrix must be 1 plane.");
		return;
	}	
	if (info.type != _jit_sym_float32)
	{
		error("2up.jit.centroids: sorry, float32 matrix needed.");
		return;
	}
	
	height = info.dim[1];
	width = info.dim[0];
	x->in_rowbytes = info.dimstride[1];
	
	// allocate image map
	if ((width != x->width) || (height != x->height))
	{
		free_map(x);
		
		if (!(x->mpc_map = (e_pixdata *)jit_getbytes(width*height)))
		{
			error("2up.jit.centroids: couldn't make image map!");
			return;
		}
		else
		{
			x->width = width; x->height = height;
		}
	}
	
	// clear map
	for (i=0; i<width*height; i++)
	{
		x->mpc_map[i] = 0;
	}

	// set direction ptrs to less intensity for neighboring pixels.
	// PIX_DOWN means that pixel in down direction has less intensity, etc. 

	for (i=0; i < height; i++)
	{
		firstrow = (i==0);
		lastrow = (i==height-1);
		pf_data = ((float *)(x->p_in_data + i*x->in_rowbytes));
		p_map = x->mpc_map + i*width;
		if (!firstrow)
		{
			pf_data_prev = pf_data - x->width;
			p_map_prev = x->mpc_map + (i-1)*width;
		}
		if (!lastrow)
		{
			pf_data_next = pf_data + x->width;
			p_map_next = x->mpc_map + (i+1)*width;
		}
		
		for (j=0; j < width; j++)
		{		
			firstcol = (j==0);
			lastcol = (j==width-1);
			pixel = pf_data[j];

			// right
			if (!lastcol)
			{
				if (pf_data[j+1] >= pixel)
					p_map[j+1] |= PIX_LEFT;
			}
			else
			{
				p_map[j] |= PIX_RIGHT;
			}
			// up
			if (!firstrow)
			{
				if (pf_data_prev[j] >= pixel)
					p_map_prev[j] |= PIX_DOWN;
			}
			else
			{
				p_map[j] |= PIX_UP;
			}
			// left
			if (!firstcol)
			{
				if (pf_data[j-1] >= pixel)
					p_map[j-1] |= PIX_RIGHT;
			}
			else
			{
				p_map[j] |= PIX_LEFT;
			}
			// down
			if (!lastrow)
			{
				if (pf_data_next[j] >= pixel)
					p_map_next[j] |= PIX_UP;
			}
			else
			{
				p_map[j] |= PIX_DOWN;
			} 
		}
	}
	
	gather_and_report_centroids(x);
}

void gather_and_report_centroids(t_jit_centroids *x)
{
	Atom av[6];
	int i, j, a;
	unsigned char * p_map;
	t_centroid_info			temp_centroid;
	int new_centroids = 0;
	int new_centroids_unculled = 0;
	float fx, fy, fp, rm, dist, min_dist;
	
	for (i=0;i<MAX_CENTROIDS;i++)
	{
		x->was_centroid[i] = 0;
	}
	
	// gathering from peaks collects pixels into new centroids
	for (i=0; i< x->height; i++)
	{
		p_map = x->mpc_map + i*x->width;
		for (j=0; j< x->width; j++)
		{
			// if peak
			if (p_map[j] == PIX_ALL)
			{
				// zero temp
				temp_centroid = g_zero_centroid;
				// gather
				gather_centroid(x, i, j, &temp_centroid);
				// if big enough, calc x and y and add to list
				if (temp_centroid.fp_sum > x->threshold)
				{
					x->p_new_centroids[new_centroids_unculled] = jit_getbytes(sizeof(t_centroid_info));
					*(x->p_new_centroids[new_centroids_unculled]) = temp_centroid;
					
					rm = 1.0 / (x->p_new_centroids[new_centroids_unculled]->fp_sum);
					fx = (float)x->p_new_centroids[new_centroids_unculled]->x_sum * rm;
					fy = (float)x->p_new_centroids[new_centroids_unculled]->y_sum * rm;
					x->p_new_centroids[new_centroids_unculled]->fx = fx;
					x->p_new_centroids[new_centroids_unculled]->fy = fy;
					new_centroids_unculled++;
					if (new_centroids_unculled >= MAX_POSSIBLE_CENTROIDS) goto done;
				}
			}
		}
	}
	
done:
	
//	post ("new_centroids_unculled: %d", new_centroids_unculled);
	
	new_centroids = MIN(x->max_centroids, new_centroids_unculled);

	// sort by intensity, cull to x->max_centroids
	for (i=0; i<new_centroids; i++)
	{
		for (j=i; j<new_centroids_unculled; j++)
		{
			if (x->p_new_centroids[j]->fp_sum > x->p_new_centroids[i]->fp_sum)
			{
				temp_centroid = *(x->p_new_centroids[j]);
				*(x->p_new_centroids[j]) = *(x->p_new_centroids[i]);
				*(x->p_new_centroids[i]) = temp_centroid;
			}
		}
	}
	
	// match current centroids up with new ones
	if (x->match)
	{
			
		// clear matches
		for (i=0; i<MAX_CENTROIDS; i++)
		{
			x->matchfwd[i] = -1;
			x->matchback[i] = -1;
		}
		
		// for all slots
		for (i=0; i<MAX_CENTROIDS; i++)
		{
			float h, v;
			// if there is an existing centroid in slot i
			if (x->p_c_info[i])
			{
				min_dist = 16384.;
				
				// for all new centroids
				for(a=0; a<new_centroids; a++) 
				{
					// get distance
					h = fabs(x->p_c_info[i]->fx - x->p_new_centroids[a]->fx);
					v = fabs(x->p_c_info[i]->fy - x->p_new_centroids[a]->fy);
					dist = sqrt(h*h + v*v);
					if (dist < min_dist)
					{
						// mark current centroid a as closest
						min_dist = dist;
						x->matchfwd[i] = a;	
					}
				}
			}
		}
		
		// for all new centroids i, find min dist to an existing centroid.
		// if no existing centroids, matchback[i] will remain -1.
		for (i=0; i<new_centroids; i++)
		{
			min_dist = 16384.;
			
			// for all slots
			for(a=0; a<MAX_CENTROIDS; a++) 
			{
				// if there is an existing centroid in slot a
				if (x->p_c_info[a])
				{
					// get city-block distance
					dist = fabs(x->p_c_info[a]->fx - x->p_new_centroids[i]->fx) + fabs(x->p_c_info[a]->fy - x->p_new_centroids[i]->fy);
					if (dist < min_dist)
					{
						// mark current centroid a as closest
						min_dist = dist;
						x->matchback[i] = a;	
					}
				}
			}
		}

		// find new centroids which are not buddies with current.  
		// if close to the real buddy, delete the new one. 
		if ((new_centroids > x->curr_centroids) && (new_centroids > 1))
		{
			for (i=0; i<new_centroids; i++)
			{		
				if (x->matchback[i] >= 0)
				{
					if (x->matchfwd[x->matchback[i]] != i)
					{
						float dx, dy;
						dx = fabs(x->p_new_centroids[x->matchfwd[x->matchback[i]]]->fx - x->p_new_centroids[i]->fx);
						dy = fabs(x->p_new_centroids[x->matchfwd[x->matchback[i]]]->fy - x->p_new_centroids[i]->fy);
						dist = sqrt(dx*dx + dy*dy);
		//post ("dist:%f", dist);
						if (dist < 3.) 
						{
		//post ("**deleting");
							// replace with last
							if (i < new_centroids-1)
							{
								*(x->p_new_centroids[i]) = *(x->p_new_centroids[new_centroids-1]);
							}
							// fix match tables
							x->matchback[new_centroids-1] = -1;
							for (a=0; a<MAX_CENTROIDS; a++)
							{
								if (x->matchfwd[a] == i) x->matchfwd[a] = -1;
								if (x->matchfwd[a] == new_centroids-1) x->matchfwd[a] = i;
							}
							new_centroids--;
							break;
						}
					}
				}
			}
		}

		// free current centroids
		for (i=0; i<MAX_CENTROIDS; i++)
		{
			if (x->p_c_info[i])
			{
				x->was_centroid[i] = TRUE;
				jit_freebytes(x->p_c_info[i], sizeof(t_centroid_info));
				x->p_c_info[i] = 0;
			}
		}
		
		// write new buddies
		for (i=0; i<new_centroids; i++)
		{
			if (x->matchback[i] >= 0)
			{
				// buddy?
				if (x->matchfwd[x->matchback[i]] == i)
				{
	//	post ("%d matched %d", i, matchback[i]);
					if (x->p_c_info[x->matchback[i]] = jit_getbytes(sizeof(t_centroid_info)))
						*x->p_c_info[x->matchback[i]] = *(x->p_new_centroids[i]);
					else
					{
						error("2up.jit.centroids: out of memory");
						return;
					}
				}
				else
				{
					x->matchback[i] = -1;
				}
			}
		}

		
		// find places for surviving non-matching
		for (i=0; i<new_centroids; i++)
		{
			if (x->matchback[i] == -1)
			{
	//post ("%d unmatched...", i);
				for (a=0; a<MAX_CENTROIDS; a++)
				{
					if (x->p_c_info[a] == 0)
					{
						if (x->p_c_info[a] = jit_getbytes(sizeof(t_centroid_info)))
						{
	//post ("found room at %d", a);
							*x->p_c_info[a] = *(x->p_new_centroids[i]);
						//	x->matchback[i] = a;	// store matchback for coloring
							break;
						}
						else
						{
							error("2up.jit.centroids: out of memory");
							return;
						}
					}
				}
			}
		}
	}
	else	// no matching, just write in order. 
	{
		// free current centroids
		for (i=0; i<MAX_CENTROIDS; i++)
		{
			if (x->p_c_info[i])
			{
				x->was_centroid[i] = TRUE;
				jit_freebytes(x->p_c_info[i], sizeof(t_centroid_info));
				x->p_c_info[i] = 0;
			}
		}
		
		for (i=0; i<new_centroids; i++)
		{
			if (x->p_c_info[i] = jit_getbytes(sizeof(t_centroid_info)))
			{
				*x->p_c_info[i] = *x->p_new_centroids[i];
			}
			else
			{
				error("2up.jit.centroids: out of memory");
				return;
			}
		}
	}

	// write "note-offs" for expired
	if (new_centroids < x->curr_centroids)
	{
		for (i=0; i<MAX_CENTROIDS; i++)
		{
			// did it turn off?
			if ((x->was_centroid[i]) && (!x->p_c_info[i]))
			{
				jit_atom_setlong(&av[0], i); 
				jit_atom_setfloat(&av[1], 0.); 
				jit_atom_setfloat(&av[2], 0.); 
				jit_atom_setfloat(&av[3], 0.); 
				outlet_list(x->listout, 0L, 4, av);
			}
		}
	}
	


// TEMP
//if (x->curr_centroids != new_centroids)
//post ("%d -> %d", x->curr_centroids, new_centroids);
	
	x->curr_centroids = new_centroids;
	
	// report each as list of index, x, y, pressure.
	for (i=0; i<MAX_CENTROIDS; i++)
	{
		if (x->p_c_info[i])
		{
			fp = (x->p_c_info[i])->fp_sum;
			fx = (x->p_c_info[i])->fx;
			fy = (x->p_c_info[i])->fy;
			
			jit_atom_setlong(&av[0], i); 
			jit_atom_setfloat(&av[1], fx / x->width); 
			jit_atom_setfloat(&av[2], fy / x->height); 
			jit_atom_setfloat(&av[3], fp); 
			outlet_list(x->listout, 0L, 4, av);
		}	
	}
	
	// free temp centroids
	for (i=0; i<new_centroids_unculled; i++)
	{
		if (x->p_new_centroids[i])
		{
			jit_freebytes(x->p_new_centroids[i], sizeof(t_centroid_info));
			x->p_new_centroids[i] = 0;
		}
	}
}


void gather_centroid(t_jit_centroids *x, int i, int j, t_centroid_info * c)
{
	register float h;				// pressure
	register float * pf_data;		// ptr to float data row
	unsigned char * p_map;
	int firstrow, lastrow, firstcol, lastcol;
		
	p_map = x->mpc_map + i*x->width;
	pf_data = ((float *)(x->p_in_data + i*x->in_rowbytes));
	h = pf_data[j];
	
	if (p_map[j] & PIX_DONE)
	{
		return;
	}
	
	// add pixel to centroid and mark as read
	c->fp_sum += h;
	c->x_sum += h*j * ((float)x->width+ 1.) / (float)x->width;
	c->y_sum += h*i * ((float)x->height+ 1.)  / (float)x->height;
	
	if (h < x->threshold)
	{
		return;
	}
	
	// mark pixel as read
	p_map[j] |= PIX_DONE;

	// recurse to any unmarked adjacent pixels of lesser intensity
	firstrow = (i==0);
	lastrow = (i==x->height-1);
	firstcol = (j==0);
	lastcol = (j==x->width-1);
	// right
	if (!lastcol)
		if (p_map[j] & PIX_RIGHT)
			gather_centroid(x, i, j+1, c);
	// up
	if (!firstrow)
		if (p_map[j] & PIX_UP)
			gather_centroid(x, i-1, j, c);
	// left
	if (!firstcol)
		if (p_map[j] & PIX_LEFT)
			gather_centroid(x, i, j-1, c);
	// down
	if (!lastrow)
		if (p_map[j] & PIX_DOWN)
			gather_centroid(x, i+1, j, c);
	
}


